@@ -167,12 +167,12 @@ transaction modes
 -----------------
 IPDB has several operating modes:
 
- * 'direct' -- any change goes immediately to the OS level
- * 'implicit' (default) -- the first change starts an implicit
-   transaction, that have to be committed
- * 'explicit' -- you have to begin() a transaction prior to
-   make any change
- * 'snapshot' -- no changes will go to the OS in any case
+    - 'direct' -- any change goes immediately to the OS level
+    - 'implicit' (default) -- the first change starts an implicit
+        transaction, that have to be committed
+    - 'explicit' -- you have to begin() a transaction prior to
+        make any change
+    - 'snapshot' -- no changes will go to the OS in any case
 
 The default is to use implicit transaction. This behaviour can
 be changed in the future, so use 'mode' argument when creating
@@ -225,8 +225,8 @@ context managers::
 On exit, the context manager will authomatically `commit()` the
 transaction.
 
-interface creation
-------------------
+create interfaces
+-----------------
 
 IPDB can also create interfaces::
 
@@ -235,15 +235,14 @@ IPDB can also create interfaces::
         i.add_port(ip.interfaces.eth2)
         i.add_ip('10.0.0.1/24')  # the same as i.add_ip('10.0.0.1', 24)
 
-Right now IPDB supports creation of `dummy`, `bond`, `bridge`
-and `vlan` interfaces. VLAN creation requires also `link` and
-`vlan_id` parameters, see example scripts.
+IPDB supports many interface types, see docs below for the
+`IPDB.create()` method.
 
 routing management
 ------------------
 
 IPDB has a simple yet useful routing management interface.
-To add a route, one can use almost any syntax::
+To add a route, there is an easy to use syntax::
 
     # spec as a dictionary
     spec = {'dst': '172.16.1.0/24',
@@ -281,6 +280,92 @@ To access and change the routes, one can use notations as follows::
     # list automatic routes keys
     print(ip.routes.tables[255].keys())
 
+**Route specs**
+
+It is important to understand, that routing tables in IPDB
+are lists, not dicts. It is still possible to use a dict syntax
+to retrieve records, but under the hood the tables are lists.
+
+To retrieve or create routes one should use route specs. The
+simplest case is to retrieve one route::
+
+    # get route by prefix
+    ip.routes['172.16.1.0/24']
+
+    # get route by a special name
+    ip.routes['default']
+
+If there are more than one route that matches the spec, only
+the first one will be retrieved. One should iterate all the
+records and filter by a key to retrieve all matches::
+
+    # only one route will be retrieved
+    ip.routes['fe80::/64']
+
+    # get all routes by this prefix
+    [ x for x in ip.routes if x['dst'] == 'fe80::/64' ]
+
+It is possible to use dicts as specs::
+
+    ip.routes[{'dst': '172.16.0.0/16',
+               'oif': 2}]
+
+The dict is just the same as a route representation in the
+records list.
+
+**Route metrics**
+
+A special object is dedicated to route metrics, one can access it
+via `route.metrics` or `route['metrics']`::
+
+    # these two statements are equal:
+    with ip.routes['172.16.1.0/24'] as route:
+        route['metrics']['mtu'] = 1400
+
+    with ip.routes['172.16.1.0/24'] as route:
+        route.metrics.mtu = 1400
+
+Possible metrics are defined in `rtmsg.py:rtmsg.metrics`, e.g.
+`RTAX_HOPLIMIT` means `hoplimit` metric etc.
+
+**Multipath routing**
+
+Multipath nexthops are managed via `route.add_nh()` and `route.del_nh()`
+methods. They are available to review via `route.multipath`, but one
+should not directly add/remove/modify nexthops `route.multipath`, as
+the changes will not be committed correctly.
+
+To create a multipath route::
+
+    ip.routes.add({'dst': '172.16.232.0/24',
+                   'multipath': [{'gateway': '172.16.231.2',
+                                  'hops': 2},
+                                 {'gateway': '172.16.231.3',
+                                  'hops': 1},
+                                 {'gateway': '172.16.231.4'}]}).commit()
+
+To change a multipath route::
+
+    with ip.routes['172.16.232.0/24'] as r:
+        r.add_nh({'gateway': '172.16.231.5'})
+        r.del_nh({'gateway': '172.16.231.4'})
+
+**On multipath hops**
+
+The `iproute2` tool uses `weigth` instead of `hops`. The weight
+is number of hops + 1, so when one creates a nexthop with `hops == 2`,
+the `iproute2` utility will show `weight 3`.
+
+But the Linux kernel uses `rtnh_hops`, and the `pyroute2` library
+uses here no implications, directly mapping the kernel provided value.
+
+**Multipath default routes**
+
+Deprecation notice: *As of the merge of kill_rtcache into the kernel,
+and it's release in ~3.6, weighted default routes no longer work*.
+Please refer to
+https://github.com/svinota/pyroute2/issues/171#issuecomment-149297244
+for details.
 
 performance issues
 ------------------
@@ -296,7 +381,6 @@ after some delay.
 classes
 -------
 '''
-import sys
 import atexit
 import logging
 import traceback
@@ -304,16 +388,19 @@ import threading
 
 from socket import AF_INET
 from socket import AF_INET6
+from pyroute2 import config
 from pyroute2.common import Dotkeys
+from pyroute2.common import View
+from pyroute2.common import basestring
+from pyroute2.common import uuid32
 from pyroute2.iproute import IPRoute
 from pyroute2.netlink.rtnl import RTM_GETLINK
 from pyroute2.ipdb.common import CreateException
 from pyroute2.ipdb.interface import Interface
 from pyroute2.ipdb.linkedset import LinkedSet
 from pyroute2.ipdb.linkedset import IPaddrSet
-from pyroute2.ipdb.common import compat
 from pyroute2.ipdb.common import SYNC_TIMEOUT
-from pyroute2.ipdb.route import RoutingTables
+from pyroute2.ipdb.route import RoutingTableSet
 
 
 def get_addr_nla(msg):
@@ -321,7 +408,7 @@ def get_addr_nla(msg):
     Utility function to get NLA, containing the interface
     address.
 
-    Incosistency in Linux IP addressing scheme is that
+    Inconsistency in Linux IP addressing scheme is that
     IPv4 uses IFA_LOCAL to store interface's ip address,
     and IPv6 uses for the same IFA_ADDRESS.
 
@@ -329,10 +416,10 @@ def get_addr_nla(msg):
     tunneling endpoint.
 
     Args:
-        * msg (nlmsg): RTM\_.*ADDR message
+        - msg (nlmsg): RTM\_.*ADDR message
 
     Returns:
-        * nla (nla): IFA_LOCAL for IPv4 and IFA_ADDRESS for IPv6
+        - nla (nla): IFA_LOCAL for IPv4 and IFA_ADDRESS for IPv6
     '''
     nla = None
     if msg['family'] == AF_INET:
@@ -345,6 +432,7 @@ def get_addr_nla(msg):
 class Watchdog(object):
     def __init__(self, ipdb, action, kwarg):
         self.event = threading.Event()
+        self.is_set = False
         self.ipdb = ipdb
 
         def cb(ipdb, msg, _action):
@@ -355,17 +443,20 @@ class Watchdog(object):
                 if (msg.get(key, None) != kwarg[key]) and \
                         (msg.get_attr(msg.name2nla(key)) != kwarg[key]):
                     return
+
+            self.is_set = True
             self.event.set()
         self.cb = cb
         # register callback prior to other things
-        self.ipdb.register_callback(self.cb)
+        self.uuid = self.ipdb.register_callback(self.cb)
 
     def wait(self, timeout=SYNC_TIMEOUT):
-        self.event.wait(timeout=timeout)
+        ret = self.event.wait(timeout=timeout)
         self.cancel()
+        return ret
 
     def cancel(self):
-        self.ipdb.unregister_callback(self.cb)
+        self.ipdb.unregister_callback(self.uuid)
 
 
 class IPDB(object):
@@ -376,23 +467,32 @@ class IPDB(object):
     '''
 
     def __init__(self, nl=None, mode='implicit',
-                 restart_on_error=None):
+                 restart_on_error=None, nl_async=None,
+                 debug=False, ignore_rtables=None):
         '''
         Parameters:
-            * nl -- IPRoute() reference
-            * mode -- (implicit, explicit, direct)
-            * iclass -- the interface class type
+            - nl -- IPRoute() reference
+            - mode -- (implicit, explicit, direct)
+            - iclass -- the interface class type
 
         If you do not provide iproute instance, ipdb will
         start it automatically.
         '''
         self.mode = mode
+        self.debug = debug
+        if isinstance(ignore_rtables, int):
+            self._ignore_rtables = [ignore_rtables, ]
+        elif isinstance(ignore_rtables, (list, tuple, set)):
+            self._ignore_rtables = ignore_rtables
+        else:
+            self._ignore_rtables = []
         self.iclass = Interface
+        self._nl_async = config.ipdb_nl_async if nl_async is None else True
         self._stop = False
         # see also 'register_callback'
-        self._post_callbacks = []
-        self._pre_callbacks = []
-        self._cb_threads = set()
+        self._post_callbacks = {}
+        self._pre_callbacks = {}
+        self._cb_threads = {}
 
         # locks and events
         self._links_event = threading.Event()
@@ -406,8 +506,7 @@ class IPDB(object):
 
         # start monitoring thread
         self._mthread = threading.Thread(target=self.serve_forever)
-        if hasattr(sys, 'ps1'):
-            self._mthread.setDaemon(True)
+        self._mthread.setDaemon(True)
         self._mthread.start()
         #
         atexit.register(self.release)
@@ -424,29 +523,40 @@ class IPDB(object):
         from scratch. Can be used when sync is lost.
         '''
         self.nl = nl or IPRoute()
-        self.nl.monitor = True
-        self.nl.bind(async=True)
 
         # resolvers
         self.interfaces = Dotkeys()
-        self.routes = RoutingTables(ipdb=self)
-        self.by_name = Dotkeys()
-        self.by_index = Dotkeys()
+        self.routes = RoutingTableSet(ipdb=self,
+                                      ignore_rtables=self._ignore_rtables)
+        self.by_name = View(src=self.interfaces,
+                            constraint=lambda k, v: isinstance(k, basestring))
+        self.by_index = View(src=self.interfaces,
+                             constraint=lambda k, v: isinstance(k, int))
 
         # caches
         self.ipaddr = {}
-        self.neighbors = {}
+        self.neighbours = {}
 
-        # load information
-        links = self.nl.get_links()
-        for link in links:
-            self.device_put(link, skip_slaves=True)
-        for link in links:
-            self.update_slaves(link)
-        self.update_addr(self.nl.get_addr())
-        self.update_neighbors(self.nl.get_neighbors())
-        routes = self.nl.get_routes()
-        self.update_routes(routes)
+        try:
+            self.nl.bind(async=self._nl_async)
+            # load information
+            links = self.nl.get_links()
+            for link in links:
+                self.device_put(link, skip_slaves=True)
+            for link in links:
+                self.update_slaves(link)
+            self.update_addr(self.nl.get_addr())
+            self.update_neighbours(self.nl.get_neighbours())
+            routes4 = self.nl.get_routes(family=AF_INET)
+            routes6 = self.nl.get_routes(family=AF_INET6)
+            self.update_routes(routes4)
+            self.update_routes(routes6)
+        except Exception as e:
+            try:
+                self.nl.close()
+            except:
+                pass
+            raise e
 
     def register_callback(self, callback, mode='post'):
         '''
@@ -482,18 +592,20 @@ class IPDB(object):
         ...
 
         The routine, `register_callback()`, takes two arguments:
-        1. callback function
-        2. mode (optional, default="post")
+            - callback function
+            - mode (optional, default="post")
 
         The callback should be a routine, that accepts three
         arguments::
 
             cb(ipdb, msg, action)
 
-        1. ipdb is a reference to IPDB instance, that invokes
-           the callback.
-        2. msg is a message arrived
-        3. action is just a msg['event'] field
+        Arguments are:
+
+            - **ipdb** is a reference to IPDB instance, that invokes
+                the callback.
+            - **msg** is a message arrived
+            - **action** is just a msg['event'] field
 
         E.g., to work on a new interface, you should catch
         action == 'RTM_NEWLINK' and with the interface index
@@ -509,23 +621,29 @@ class IPDB(object):
                 callback(*argv, **kwarg)
 
         safe.hook = callback
-        if mode == 'post':
-            self._post_callbacks.append(safe)
-        elif mode == 'pre':
-            self._pre_callbacks.append(safe)
+        safe.lock = lock
+        safe.uuid = uuid32()
 
-    def unregister_callback(self, callback, mode='post'):
+        if mode == 'post':
+            self._post_callbacks[safe.uuid] = safe
+        elif mode == 'pre':
+            self._pre_callbacks[safe.uuid] = safe
+        return safe.uuid
+
+    def unregister_callback(self, cuid, mode='post'):
         if mode == 'post':
             cbchain = self._post_callbacks
         elif mode == 'pre':
             cbchain = self._pre_callbacks
         else:
             raise KeyError('Unknown callback mode')
-        for cb in tuple(cbchain):
-            if callback == cb.hook:
-                for t in tuple(self._cb_threads):
-                    t.join(3)
-                return cbchain.pop(cbchain.index(cb))
+        safe = cbchain[cuid]
+        with safe.lock:
+            cbchain.pop(cuid)
+        for t in tuple(self._cb_threads.get(cuid, ())):
+            t.join(3)
+        ret = self._cb_threads.get(cuid, ())
+        return ret
 
     def release(self):
         '''
@@ -539,34 +657,63 @@ class IPDB(object):
         is enough time to sync the state. But for the scripts
         the `release()` call is required.
         '''
-        with self._shutdown_lock:
-            if self._stop:
-                return
+        with self.exclusive:
+            with self._shutdown_lock:
+                if self._stop:
+                    return
 
-            self._stop = True
-            try:
-                self.nl.put({'index': 1}, RTM_GETLINK)
-                self._mthread.join()
-            except Exception:
-                # Just give up.
-                # We can not handle this case
-                pass
-            self.nl.close()
+                self._stop = True
+                # collect all the callbacks
+                for cuid in tuple(self._cb_threads):
+                    for t in tuple(self._cb_threads[cuid]):
+                        t.join()
+                # terminate the main loop
+                try:
+                    for t in range(3):
+                        self.nl.put({'index': 1}, RTM_GETLINK)
+                        self._mthread.join(t)
+                        if not self._mthread.is_alive():
+                            break
+                except Exception:
+                    # Just give up.
+                    # We can not handle this case
+                    pass
+                self.nl.close()
+                self.nl = None
+
+                # flush all the objects
+                # -- interfaces
+                for (key, dev) in self.by_name.items():
+                    self.detach(key, dev['index'], dev.nlmsg)
+                # -- routes
+                for key in tuple(self.routes.tables.keys()):
+                    del self.routes.tables[key]
+                self.routes.tables[254] = None
+                # -- ipaddr
+                for key in tuple(self.ipaddr.keys()):
+                    del self.ipaddr[key]
+                # -- neighbours
+                for key in tuple(self.neighbours.keys()):
+                    del self.neighbours[key]
 
     def create(self, kind, ifname, reuse=False, **kwarg):
         '''
         Create an interface. Arguments 'kind' and 'ifname' are
         required.
 
-        * kind -- interface type, can be of:
-          * bridge
-          * bond
-          * vlan
-          * tun
-          * dummy
-          * veth
-        * ifname -- interface name
-        * reuse -- if such interface exists, return it anyway
+            - kind — interface type, can be of:
+                - bridge
+                - bond
+                - vlan
+                - tun
+                - dummy
+                - veth
+                - macvlan
+                - macvtap
+                - gre
+                - team
+            - ifname — interface name
+            - reuse — if such interface exists, return it anyway
 
         Different interface kinds can require different
         arguments for creation.
@@ -583,6 +730,60 @@ class IPDB(object):
         The code above creates two interfaces, `v1p0` and `v1p1`, and
         adds two addresses to `v1p0`.
 
+        ► **macvlan**
+
+        Macvlan interfaces act like VLANs within OS. The macvlan driver
+        provides an ability to add several MAC addresses on one interface,
+        where every MAC address is reflected with a virtual interface in
+        the system.
+
+        In some setups macvlan interfaces can replace bridge interfaces,
+        providing more simple and at the same time high-performance
+        solution::
+
+            ip.create(ifname='mvlan0',
+                      kind='macvlan',
+                      link=ip.interfaces.em1,
+                      macvlan_mode='private').commit()
+
+        Several macvlan modes are available: 'private', 'vepa', 'bridge',
+        'passthru'. Ususally the default is 'vepa'.
+
+        ► **macvtap**
+
+        Almost the same as macvlan, but creates also a character tap device::
+
+            ip.create(ifname='mvtap0',
+                      kind='macvtap',
+                      link=ip.interfaces.em1,
+                      macvtap_mode='vepa').commit()
+
+        Will create a device file `"/dev/tap%s" % ip.interfaces.mvtap0.index`
+
+        ► **gre**
+
+        Create GRE tunnel::
+
+            with ip.create(ifname='grex',
+                           kind='gre',
+                           gre_local='172.16.0.1',
+                           gre_remote='172.16.0.101',
+                           gre_ttl=16) as i:
+                i.add_ip('192.168.0.1/24')
+                i.up()
+
+        The keyed GRE requires explicit iflags/oflags specification::
+
+            ip.create(ifname='grex',
+                      kind='gre',
+                      gre_local='172.16.0.1',
+                      gre_remote='172.16.0.101',
+                      gre_ttl=16,
+                      gre_ikey=10,
+                      gre_okey=10,
+                      gre_iflags=32,
+                      gre_oflags=32).commit()
+
         ► **vlan**
 
         VLAN interfaces require additional parameters, `vlan_id` and
@@ -601,59 +802,126 @@ class IPDB(object):
         The `link` parameter should be either integer, interface id, or
         an interface object. VLAN id must be integer.
 
+        ► **vxlan**
+
+        VXLAN interfaces are like VLAN ones, but require a bit more
+        parameters::
+
+            ip.create(ifname='vx101',
+                      kind='vxlan',
+                      vxlan_link=ip.interfaces.eth0,
+                      vxlan_id=101,
+                      vxlan_group='239.1.1.1',
+                      vxlan_ttl=16)
+
+        All possible vxlan parameters are listed in the module
+        `pyroute2.netlink.rtnl.ifinfmsg:... vxlan_data`.
+
         ► **tuntap**
 
         Possible `tuntap` keywords:
 
-            * `mode` — "tun" or "tap"
-            * `uid` — integer
-            * `gid` — integer
-            * `ifr` — dict of tuntap flags (see tuntapmsg.py)
+            - `mode` — "tun" or "tap"
+            - `uid` — integer
+            - `gid` — integer
+            - `ifr` — dict of tuntap flags (see tuntapmsg.py)
         '''
         with self.exclusive:
             # check for existing interface
             if ifname in self.interfaces:
-                if self.interfaces[ifname]._flicker or reuse:
+                if (self.interfaces[ifname]['ipdb_scope'] == 'shadow') \
+                        or reuse:
                     device = self.interfaces[ifname]
-                    device._flicker = False
+                    kwarg['kind'] = kind
+                    device.load_dict(kwarg)
+                    device.set_item('ipdb_scope', 'create')
                 else:
                     raise CreateException("interface %s exists" %
                                           ifname)
             else:
                 device = \
-                    self.by_name[ifname] = \
                     self.interfaces[ifname] = \
                     self.iclass(ipdb=self, mode='snapshot')
                 device.update(kwarg)
                 if isinstance(kwarg.get('link', None), Interface):
                     device['link'] = kwarg['link']['index']
+                if isinstance(kwarg.get('vxlan_link', None), Interface):
+                    device['vxlan_link'] = kwarg['vxlan_link']['index']
                 device['kind'] = kind
                 device['index'] = kwarg.get('index', 0)
                 device['ifname'] = ifname
+                device['ipdb_scope'] = 'create'
                 device._mode = self.mode
-            device.begin()
-            return device
+            tid = device.begin()
+        #
+        # All the device methods are handled via `transactional.update()`
+        # except of the very creation.
+        #
+        # Commit the changes in the 'direct' mode, since this call is not
+        # decorated.
+        if self.mode == 'direct':
+            device.commit(tid)
+        return device
+
+    def commit(self, transactions=None, rollback=False):
+        # what to commit: either from transactions argument, or from
+        # started transactions on existing objects
+        if transactions is None:
+            # collect interface transactions
+            txlist = [(x, x.last()) for x in self.by_name.values() if x._tids]
+            # collect route transactions
+            for table in self.routes.tables.keys():
+                txlist.extend([(x, x.last()) for x in
+                               self.routes.tables[table]
+                               if x._tids])
+            txlist = sorted(txlist,
+                            key=lambda x: x[1]['ipdb_priority'],
+                            reverse=True)
+            transactions = txlist
+
+        snapshots = []
+        removed = []
+
+        try:
+            for (target, tx) in transactions:
+                if target['ipdb_scope'] == 'detached':
+                    continue
+                if tx['ipdb_scope'] == 'remove':
+                    tx['ipdb_scope'] = 'shadow'
+                    removed.append((target, tx))
+                if not rollback:
+                    s = (target, target.pick(detached=True))
+                    snapshots.append(s)
+                target.commit(transaction=tx, rollback=rollback)
+        except Exception:
+            if not rollback:
+                self.fallen = transactions
+                self.commit(transactions=snapshots, rollback=True)
+            raise
+        else:
+            if not rollback:
+                for (target, tx) in removed:
+                    target['ipdb_scope'] = 'detached'
+                    target.detach()
+        finally:
+            if not rollback:
+                for (target, tx) in transactions:
+                    target.drop(tx)
 
     def device_del(self, msg):
-        # check for flicker devices
-        if (msg.get('index', None) in self.interfaces) and \
-                self.interfaces[msg['index']]._flicker:
+        target = self.interfaces.get(msg['index'])
+        if target is None:
+            return
+        target.nlmsg = msg
+        # check for freezed devices
+        if getattr(target, '_freeze', None):
+            self.interfaces[msg['index']].set_item('ipdb_scope', 'shadow')
+            return
+        # check for locked devices
+        if target.get('ipdb_scope') in ('locked', 'shadow'):
             self.interfaces[msg['index']].sync()
             return
-        try:
-            self.update_slaves(msg)
-            if msg['change'] == 0xffffffff:
-                # FIXME catch exception
-                ifname = self.interfaces[msg['index']]['ifname']
-                self.interfaces[msg['index']].sync()
-                del self.by_name[ifname]
-                del self.by_index[msg['index']]
-                del self.interfaces[ifname]
-                del self.interfaces[msg['index']]
-                del self.ipaddr[msg['index']]
-                del self.neighbors[msg['index']]
-        except KeyError:
-            pass
+        self.detach(None, msg['index'], msg)
 
     def device_put(self, msg, skip_slaves=False):
         # check, if a record exists
@@ -667,29 +935,22 @@ class IPDB(object):
         if ((index not in self.interfaces) and
                 (ifname not in self.interfaces)):
             # scenario #1, new interface
-            if compat.fix_check_link(self.nl, index):
-                return
             device = \
-                self.by_index[index] = \
                 self.interfaces[index] = \
-                self.interfaces[ifname] = \
-                self.by_name[ifname] = self.iclass(ipdb=self)
+                self.interfaces[ifname] = self.iclass(ipdb=self)
         elif ((index not in self.interfaces) and
                 (ifname in self.interfaces)):
             # scenario #2, index change
             old_index = self.interfaces[ifname]['index']
-            device = \
-                self.interfaces[index] = \
-                self.by_index[index] = self.interfaces[ifname]
+            device = self.interfaces[index] = self.interfaces[ifname]
             if old_index in self.interfaces:
                 del self.interfaces[old_index]
-                del self.by_index[old_index]
             if old_index in self.ipaddr:
                 self.ipaddr[index] = self.ipaddr[old_index]
                 del self.ipaddr[old_index]
-            if old_index in self.neighbors:
-                self.neighbors[index] = self.neighbors[old_index]
-                del self.neighbors[old_index]
+            if old_index in self.neighbours:
+                self.neighbours[index] = self.neighbours[old_index]
+                del self.neighbours[old_index]
         else:
             # scenario #3, interface rename
             # scenario #4, assume rename
@@ -697,41 +958,73 @@ class IPDB(object):
             if old_name != ifname:
                 # unlink old name
                 del self.interfaces[old_name]
-                del self.by_name[old_name]
-            device = \
-                self.interfaces[ifname] = \
-                self.by_name[ifname] = self.interfaces[index]
+            device = self.interfaces[ifname] = self.interfaces[index]
 
         if index not in self.ipaddr:
             # for interfaces, created by IPDB
             self.ipaddr[index] = IPaddrSet()
 
-        if index not in self.neighbors:
-            self.neighbors[index] = LinkedSet()
+        if index not in self.neighbours:
+            self.neighbours[index] = LinkedSet()
 
         device.load_netlink(msg)
 
         if not skip_slaves:
             self.update_slaves(msg)
 
-    def detach(self, item):
+    def detach(self, name, idx, msg=None):
         with self.exclusive:
-            if item in self.interfaces:
-                del self.interfaces[item]
-            if item in self.by_name:
-                del self.by_name[item]
-            if item in self.by_index:
-                del self.by_index[item]
+            if msg is not None:
+                try:
+                    self.update_slaves(msg)
+                except KeyError:
+                    pass
+                if msg['event'] == 'RTM_DELLINK' and \
+                        msg['change'] != 0xffffffff:
+                    return
+            if idx is None or idx < 1:
+                target = self.interfaces[name]
+                idx = target['index']
+            else:
+                target = self.interfaces[idx]
+                name = target['ifname']
+            target.sync()
+            self.interfaces.pop(name, None)
+            self.interfaces.pop(idx, None)
+            self.ipaddr.pop(idx, None)
+            self.neighbours.pop(idx, None)
+            target.set_item('ipdb_scope', 'detached')
 
     def watchdog(self, action='RTM_NEWLINK', **kwarg):
         return Watchdog(self, action, kwarg)
 
+    def update_dev(self, dev):
+        # ignore non-system updates on devices not
+        # registered in the DB
+        if (dev['index'] not in self.interfaces) and \
+                (dev['change'] != 0xffffffff):
+            return
+        if dev['event'] == 'RTM_NEWLINK':
+            self.device_put(dev)
+        else:
+            self.device_del(dev)
+
     def update_routes(self, routes):
         for msg in routes:
             self.routes.load_netlink(msg)
 
     def _lookup_master(self, msg):
-        master = msg.get_attr('IFLA_MASTER')
+        master = None
+        # lookup for IFLA_INFO_OVS_MASTER
+        li = msg.get_attr('IFLA_LINKINFO')
+        if li:
+            master = li.get_attr('IFLA_INFO_OVS_MASTER')
+        # lookup for IFLA_MASTER
+        if master is None:
+            master = msg.get_attr('IFLA_MASTER')
+        # pls keep in mind, that in the case of IFLA_MASTER
+        # lookup is done via interface index, while in the case
+        # of IFLA_INFO_OVS_MASTER lookup is done via ifname
         return self.interfaces.get(master, None)
 
     def update_slaves(self, msg):
@@ -751,8 +1044,11 @@ class IPDB(object):
                 # masters refers to the same slave
                 for device in self.by_index:
                     if index in self.interfaces[device]['ports']:
-                        self.interfaces[device].del_port(index,
-                                                         direct=True)
+                        try:
+                            self.interfaces[device].del_port(
+                                index, direct=True)
+                        except KeyError:
+                            pass
                 master.add_port(index, direct=True)
             elif msg['event'] == 'RTM_DELLINK':
                 if index in master['ports']:
@@ -764,37 +1060,55 @@ class IPDB(object):
             # clean device from ports
             for master in self.by_index:
                 if index in self.interfaces[master]['ports']:
-                    self.interfaces[master].del_port(index,
-                                                     direct=True)
+                    try:
+                        self.interfaces[master].del_port(
+                            index, direct=True)
+                    except KeyError:
+                        pass
             master = device.if_master
             if master is not None:
                 if 'master' in device:
-                    device.del_item('master')
+                    device.set_item('master', None)
                 if (master in self.interfaces) and \
                         (msg['index'] in self.interfaces[master].ports):
-                    self.interfaces[master].del_port(msg['index'],
-                                                     direct=True)
+                    try:
+                        self.interfaces[master].del_port(
+                            msg['index'], direct=True)
+                    except KeyError:
+                        pass
 
     def update_addr(self, addrs, action='add'):
         # Update address list of an interface.
 
         for addr in addrs:
             nla = get_addr_nla(addr)
+            if self.debug:
+                raw = addr
+            else:
+                raw = {'local': addr.get_attr('IFA_LOCAL'),
+                       'broadcast': addr.get_attr('IFA_BROADCAST'),
+                       'address': addr.get_attr('IFA_ADDRESS'),
+                       'flags': addr.get_attr('IFA_FLAGS'),
+                       'prefixlen': addr.get('prefixlen')}
             if nla is not None:
                 try:
                     method = getattr(self.ipaddr[addr['index']], action)
-                    method(key=(nla, addr['prefixlen']), raw=addr)
+                    method(key=(nla, addr['prefixlen']), raw=raw)
                 except:
                     pass
 
-    def update_neighbors(self, neighs, action='add'):
+    def update_neighbours(self, neighs, action='add'):
 
         for neigh in neighs:
             nla = neigh.get_attr('NDA_DST')
+            if self.debug:
+                raw = neigh
+            else:
+                raw = {'lladdr': neigh.get_attr('NDA_LLADDR')}
             if nla is not None:
                 try:
-                    method = getattr(self.neighbors[neigh['ifindex']], action)
-                    method(key=nla, raw=neigh)
+                    method = getattr(self.neighbours[neigh['ifindex']], action)
+                    method(key=nla, raw=raw)
                 except:
                     pass
 
@@ -823,60 +1137,61 @@ class IPDB(object):
                 logging.error('Restarting IPDB instance after '
                               'error:\n%s', traceback.format_exc())
                 if self.restart_on_error:
-                    self.initdb()
+                    try:
+                        self.initdb()
+                    except:
+                        logging.error('Error restarting DB:\n%s',
+                                      traceback.format_exc())
+                        return
                     continue
                 else:
                     raise RuntimeError('Emergency shutdown')
             for msg in messages:
                 # Run pre-callbacks
                 # NOTE: pre-callbacks are synchronous
-                for cb in self._pre_callbacks:
+                for (cuid, cb) in tuple(self._pre_callbacks.items()):
                     try:
                         cb(self, msg, msg['event'])
                     except:
                         pass
 
                 with self.exclusive:
-                    if msg.get('event', None) == 'RTM_NEWLINK':
-                        self.device_put(msg)
+                    # FIXME: refactor it to a dict
+                    if msg.get('event', None) in ('RTM_NEWLINK',
+                                                  'RTM_DELLINK'):
+                        self.update_dev(msg)
                         self._links_event.set()
-                    elif msg.get('event', None) == 'RTM_DELLINK':
-                        self.device_del(msg)
                     elif msg.get('event', None) == 'RTM_NEWADDR':
                         self.update_addr([msg], 'add')
                     elif msg.get('event', None) == 'RTM_DELADDR':
                         self.update_addr([msg], 'remove')
                     elif msg.get('event', None) == 'RTM_NEWNEIGH':
-                        self.update_neighbors([msg], 'add')
+                        self.update_neighbours([msg], 'add')
                     elif msg.get('event', None) == 'RTM_DELNEIGH':
-                        self.update_neighbors([msg], 'remove')
-                    elif msg.get('event', None) == 'RTM_NEWROUTE':
+                        self.update_neighbours([msg], 'remove')
+                    elif msg.get('event', None) in ('RTM_NEWROUTE',
+                                                    'RTM_DELROUTE'):
                         self.update_routes([msg])
-                    elif msg.get('event', None) == 'RTM_DELROUTE':
-                        table = msg.get('table', 254)
-                        dst = msg.get_attr('RTA_DST', False)
-                        if not dst:
-                            key = 'default'
-                        else:
-                            key = '%s/%s' % (dst, msg.get('dst_len', 0))
-                        try:
-                            route = self.routes.tables[table][key]
-                            del self.routes.tables[table][key]
-                            route.sync()
-                        except KeyError:
-                            pass
 
                 # run post-callbacks
                 # NOTE: post-callbacks are asynchronous
-                for cb in self._post_callbacks:
+                for (cuid, cb) in tuple(self._post_callbacks.items()):
                     t = threading.Thread(name="callback %s" % (id(cb)),
                                          target=cb,
                                          args=(self, msg, msg['event']))
                     t.start()
-                    self._cb_threads.add(t)
+                    if cuid not in self._cb_threads:
+                        self._cb_threads[cuid] = set()
+                    self._cb_threads[cuid].add(t)
 
                 # occasionally join cb threads
-                for t in tuple(self._cb_threads):
-                    t.join(0)
-                    if not t.is_alive():
-                        self._cb_threads.remove(t)
+                for cuid in tuple(self._cb_threads):
+                    for t in tuple(self._cb_threads.get(cuid, ())):
+                        t.join(0)
+                        if not t.is_alive():
+                            try:
+                                self._cb_threads[cuid].remove(t)
+                            except KeyError:
+                                pass
+                            if len(self._cb_threads.get(cuid, ())) == 0:
+                                del self._cb_threads[cuid]
