@@ -5,17 +5,17 @@ Netlink basics
 General netlink packet structure::
 
     nlmsg packet:
-        + header
-        + data
+        header
+        data
 
 Generic netlink message header::
 
     nlmsg header:
-        + uint32 length
-        + uint16 type
-        + uint16 flags
-        + uint32 sequence number
-        + uint32 pid
+        uint32 length
+        uint16 type
+        uint16 flags
+        uint32 sequence number
+        uint32 pid
 
 The `length` field is the length of all the packet, including
 data and header. The `type` field is used to distinguish different
@@ -39,25 +39,25 @@ netlink messages can carry NLA (netlink attributes). NLA
 structure is as follows::
 
     NLA header:
-        + uint16 length
-        + uint16 type
+        uint16 length
+        uint16 type
     NLA data:
-        + data-specific struct
+        data-specific struct
         # optional:
-        + NLA
-        + NLA
-        + ...
+        NLA
+        NLA
+        ...
 
 So, NLA structures can be nested, forming a tree.
 
 Complete structure of a netlink packet::
 
     nlmsg header:
-        + uint32 length
-        + uint16 type
-        + uint16 flags
-        + uint32 sequence number
-        + uint32 pid
+        uint32 length
+        uint16 type
+        uint16 flags
+        uint32 sequence number
+        uint32 pid
     [ optional protocol-specific data ]
     [ optional NLA tree ]
 
@@ -68,7 +68,7 @@ Pyroute2 and netlink
 ====================
 
 packets
-+++++++
+-------
 
 To simplify the development, pyroute2 provides an easy way to
 describe packet structure. As an example, you can take the
@@ -87,7 +87,7 @@ NLA are described in the same way, but the parent class should be
 proper parent class -- it affects the header structure.
 
 fields attribute
-++++++++++++++++
+----------------
 
 The `fields` attribute describes the structure of the
 protocol-specific data. It is a tuple of tuples, where each
@@ -96,12 +96,13 @@ member contains a field name and its data format.
 Field data format should be specified as for Python `struct`
 module. E.g., ifaddrmsg structure::
 
-    ifaddrmsg structure:
-        + unsigned char ifa_family
-        + unsigned char ifa_prefixlen
-        + unsigned char ifa_flags
-        + unsigned char ifa_scope
-        + int ifa_index
+    struct ifaddrmsg {
+        __u8  ifa_family;
+        __u8  ifa_prefixlen;
+        __u8  ifa_flags;
+        __u8  ifa_scope;
+        __u32 ifa_index;
+    };
 
 should be described as follows::
 
@@ -137,7 +138,7 @@ packing doesn't work well::
 
 
 nla_map attribute
-+++++++++++++++++
+-----------------
 
 The `nla_map` attribute is a tuple of NLA descriptions. Each
 description is also a tuple in two different forms: either
@@ -165,15 +166,20 @@ are resolved in runtime.
 There are several pre-defined NLA types, that you will get with
 `nla` class:
 
-* none  # forces pyroute2 just to skip this NLA
-* uint8
-* uint16
-* uint32  # there are dedicated NLA of these types as well
-* ipaddr  # IP address, IPv4 or IPv6, depending on the socket
-* l2addr  # MAC address
-* hex     # hex dump as a string -- useful for debugging
-* cdata   # just a binary string
-* asciiz  # zero-terminated ASCII string
+    - `none` -- ignore this NLA
+    - `flag` -- boolean flag NLA (no payload; NLA exists = True)
+    - `uint8`, `uint16`, `uint32`, `uint64` -- unsigned int
+    - `be8`, `be16`, `be32`, `be64` -- big-endian unsigned int
+    - `ipaddr` -- IP address, IPv4 or IPv6
+    - `ip4addr` -- only IPv4 address type
+    - `ip6addr` -- only IPv6 address type
+    - `target` -- a univeral target (IPv4, IPv6, MPLS)
+    - `l2addr` -- MAC address
+    - `hex` -- hex dump as a string -- useful for debugging
+    - `cdata` -- a binary data
+    - `string` -- UTF-8 string
+    - `asciiz` -- zero-terminated ASCII string, no decoding
+    - `array` -- array of simple types (uint8, uint16 etc.)
 
 Please refer to `pyroute2/netlink/__init__.py` for details.
 
@@ -210,8 +216,40 @@ It is not permitted to mix these two forms in one class: you should
 use ether autogenerated type numbers (two fields tuples), or
 explicit numbers (three fields typles).
 
+array types
+-----------
+
+There are different array-like NLA types in the kernel, and
+some of them are covered by pyroute2. An array of simple type
+elements::
+
+    # declaration
+    nla_map = (('NLA_TYPE', 'array(uint8)'), ...)
+
+    # data layout
+    +======+======+----------------------------
+    | len  | type | uint8 | uint8 | uint 8 | ...
+    +======+======+----------------------------
+
+    # decoded
+    {'attrs': [['NLA_TYPE', (2, 3, 4, 5, ...)], ...], ...}
+
+An array of NLAs::
+
+    # declaration
+    nla_map = (('NLA_TYPE', '*type'), ...)
+
+    # data layout
+    +=======+=======+-----------------------+-----------------------+--
+    | len   | type* | len  | type | payload | len  | type | payload | ...
+    +=======+=======+-----------------------+-----------------------+--
+    # type* -- in that case the type is OR'ed with NLA_F_NESTED
+
+    # decoded
+    {'attrs': [['NLA_TYPE', [payload, payload, ...]], ...], ...}
+
 parsed netlink message
-++++++++++++++++++++++
+----------------------
 
 Netlink messages are represented by pyroute2 as dictionaries
 as follows::
@@ -258,8 +296,36 @@ scan event::
                 {'attrs': [['SIOCGIWSCAN',
                             '00:00:00:00:00:00:00:00:00:00:00:00']]}]]}
 
+One important detail is that NLA chain is represented as a list of
+elements `['NLA_TYPE', value]`, not as a dictionary. The reason is that
+though in the kernel *usually* NLA chain is a dictionary, the netlink
+protocol by itself doesn't require elements of each type to be unique.
+In a message there may be several NLA of the same type.
+
+encoding and decoding algo
+--------------------------
+
+The message encoding works as follows:
+
+1. Reserve space for the message header (if there is)
+2. Iterate defined `fields`, encoding values with `struct.pack()`
+3. Iterate NLA from the `attrs` field, looking up types in `nla_map`
+4. Encode the header
+
+Since every NLA is also an `nlmsg` object, there is a recursion.
+
+The decoding process is a bit simpler:
+
+1. Decode the header
+2. Iterate `fields`, decoding values with `struct.unpack()`
+3. Iterate NLA until the message ends
+
+If the `fields` attribute is an empty list, the step 2 will be skipped.
+The step 3 will be skipped in the case of the empty `nla_map`. If both
+attributes are empty lists, only the header will be encoded/decoded.
+
 create and send messages
-++++++++++++++++++++++++
+------------------------
 
 Using high-level interfaces like `IPRoute` or `IPDB`, you will never
 need to manually construct and send netlink messages. But in the case
@@ -267,10 +333,10 @@ you really need it, it is simple as well.
 
 Having a description class, like `ifaddrmsg` from above, you need to:
 
-* instantiate it
-* fill the fields
-* encode the packet
-* send the encoded data
+    - instantiate it
+    - fill the fields
+    - encode the packet
+    - send the encoded data
 
 The code::
 
@@ -322,7 +388,6 @@ Module contents:
 
 import traceback
 import logging
-import socket
 import struct
 import types
 import sys
@@ -330,6 +395,12 @@ import io
 import re
 import os
 
+from socket import inet_pton
+from socket import inet_ntop
+from socket import AF_INET
+from socket import AF_INET6
+from socket import AF_UNSPEC
+from pyroute2.common import AF_MPLS
 from pyroute2.common import hexdump
 from pyroute2.common import basestring
 
@@ -415,8 +486,9 @@ NETLINK_GENERIC = 16
 # leave room for NETLINK_DM (DM Events)
 NETLINK_SCSITRANSPORT = 18   # SCSI Transports
 
-
-NLMSG_ALIGNTO = 4
+# NLA flags
+NLA_F_NESTED = 1 << 15
+NLA_F_NET_BYTEORDER = 1 << 14
 
 
 class NetlinkError(Exception):
@@ -460,10 +532,6 @@ class NetlinkNLADecodeError(NetlinkDecodeError):
     pass
 
 
-def NLMSG_ALIGN(l):
-    return (l + NLMSG_ALIGNTO - 1) & ~ (NLMSG_ALIGNTO - 1)
-
-
 class NotInitialized(Exception):
     pass
 
@@ -517,6 +585,18 @@ IPRCMD_REMOVE = 14
 IPRCMD_DISCOVER = 15
 IPRCMD_UNREGISTER = 16
 
+SOL_NETLINK = 270
+
+NETLINK_ADD_MEMBERSHIP = 1
+NETLINK_DROP_MEMBERSHIP = 2
+NETLINK_PKTINFO = 3
+NETLINK_BROADCAST_ERROR = 4
+NETLINK_NO_ENOBUFS = 5
+NETLINK_RX_RING = 6
+NETLINK_TX_RING = 7
+
+clean_cbs = {}
+
 
 class nlmsg_base(dict):
     '''
@@ -532,9 +612,23 @@ class nlmsg_base(dict):
     fields = []                  # data field names, to build a dictionary
     header = None                # optional header class
     pack = None                  # pack pragma
+    nla_array = False
+    cell_header = None
+    align = 4
     nla_map = {}                 # NLA mapping
+    nla_flags = 0                # NLA flags
+    nla_init = None              # NLA initstring
+    value_map = {}
 
-    def __init__(self, buf=None, length=None, parent=None, debug=False):
+    def msg_align(self, l):
+        return (l + self.align - 1) & ~ (self.align - 1)
+
+    def __init__(self,
+                 buf=None,
+                 length=None,
+                 parent=None,
+                 debug=False,
+                 init=None):
         dict.__init__(self)
         for i in self.fields:
             self[i[0]] = 0  # FIXME: only for number values
@@ -544,10 +638,12 @@ class nlmsg_base(dict):
         self.parent = parent
         self.offset = 0
         self.prefix = None
+        self.nla_init = init
         self['attrs'] = []
         self['value'] = NotInitialized
         self.value = NotInitialized
         self.register_nlas()
+        self.r_value_map = dict([(x[1], x[0]) for x in self.value_map.items()])
         self.reset(buf)
         if self.header is not None:
             self['header'] = self.header(self.buf)
@@ -580,6 +676,33 @@ class nlmsg_base(dict):
         if 'header' in self:
             self['header'].buf = self.buf
 
+    def register_clean_cb(self, cb):
+        global clean_cbs
+        if self.parent is not None:
+            return self.parent.register_clean_cb(cb)
+        else:
+            # get the msg_seq -- if applicable
+            seq = self.get('header', {}).get('sequence_number', None)
+            if seq is not None and seq not in clean_cbs:
+                clean_cbs[seq] = []
+            # attach the callback
+            clean_cbs[seq].append(cb)
+
+    def unregister_clean_cb(self):
+        global clean_cbs
+        seq = self.get('header', {}).get('sequence_number', None)
+        msf = self.get('header', {}).get('flags', 0)
+        if (seq is not None) and \
+                (not msf & NLM_F_REQUEST) and \
+                seq in clean_cbs:
+            for cb in clean_cbs[seq]:
+                try:
+                    cb()
+                except:
+                    logging.error('Cleanup callback fail: %s' % (cb))
+                    logging.error(traceback.format_exc())
+            del clean_cbs[seq]
+
     def _strip_one(self, name):
         for i in tuple(self['attrs']):
             if i[0] == name:
@@ -663,7 +786,15 @@ class nlmsg_base(dict):
         if lvalue is self:
             for key in self:
                 try:
-                    assert self.get(key) == rvalue.get(key)
+                    lv = self.get(key)
+                    rv = rvalue.get(key)
+                    # this strange condition means a simple thing:
+                    # None, 0, empty container and NotInitialized in
+                    # that context should be treated as equal.
+                    if (lv != rv) and not \
+                            ((not lv or lv is NotInitialized) and
+                             (not rv or rv is NotInitialized)):
+                        return False
                 except Exception:
                     # on any error -- is not equal
                     return False
@@ -714,38 +845,7 @@ class nlmsg_base(dict):
             size += struct.calcsize(i[1])
         self.buf.seek(size, 1)
 
-    def decode(self):
-        '''
-        Decode the message. The message should have the `buf`
-        attribute initialized. e.g.::
-
-            data = sock.recv(16384)
-            msg = ifinfmsg(data)
-
-        If you want to customize the decoding process, override
-        the method, but don't forget to call parent's `decode()`::
-
-            class CustomMessage(nlmsg):
-
-                def decode(self):
-                    nlmsg.decode(self)
-                    ...  # do some custom data tuning
-        '''
-        self.offset = self.buf.tell()
-        # decode the header
-        if self.header is not None:
-            try:
-                self['header'].decode()
-                # update length from header
-                # it can not be less than 4
-                self.length = max(self['header']['length'], 4)
-                save = self.buf.tell()
-                self.buf.seek(self.offset)
-                self.raw = self.buf.read(self.length)
-                self.buf.seek(save)
-            except Exception as e:
-                raise NetlinkHeaderDecodeError(e)
-        # decode the data
+    def decode_fields(self):
         try:
             if self.pack == 'struct':
                 names = []
@@ -780,7 +880,8 @@ class nlmsg_base(dict):
                         self[name] = value[0]
                         # cut zero-byte from z-strings
                         # 0x00 -- python3; '\0' -- python2
-                        if field[1] == 'z' and self[name][-1] in (0x00, '\0'):
+                        if field[1] == 'z' and self[name][-1] \
+                                in (0x00, '\0'):
                             self[name] = self[name][:-1]
                     else:
                         if self.pack == 'struct':
@@ -798,14 +899,61 @@ class nlmsg_base(dict):
 
         except Exception as e:
             raise NetlinkDataDecodeError(e)
+
+    def decode(self):
+        '''
+        Decode the message. The message should have the `buf`
+        attribute initialized. e.g.::
+
+            data = sock.recv(16384)
+            msg = ifinfmsg(data)
+
+        If you want to customize the decoding process, override
+        the method, but don't forget to call parent's `decode()`::
+
+            class CustomMessage(nlmsg):
+
+                def decode(self):
+                    nlmsg.decode(self)
+                    ...  # do some custom data tuning
+        '''
+        self.offset = self.buf.tell()
+        # decode the header
+        if self.header is not None:
+            try:
+                self['header'].decode()
+                # update length from header
+                # it can not be less than 4
+                self.length = max(self['header']['length'], 4)
+                save = self.buf.tell()
+                self.buf.seek(self.offset)
+                self.raw = self.buf.read(self.length)
+                self.buf.seek(save)
+            except Exception as e:
+                raise NetlinkHeaderDecodeError(e)
+        # handle the array case
+        if self.nla_array:
+            self.setvalue([])
+            while self.buf.tell() < self.offset + self.length:
+                cell = type(self)(self.buf, parent=self, debug=self.debug)
+                cell.nla_array = False
+                if cell.cell_header is not None:
+                    cell.header = cell.cell_header
+                    cell['header'] = cell.cell_header(self.buf)
+                cell.decode()
+                self.value.append(cell)
+        else:
+            # decode data
+            self.decode_fields()
         # decode NLA
         try:
-            # align NLA chain start
-            self.buf.seek(NLMSG_ALIGN(self.buf.tell()))
+            self.unregister_clean_cb()
             # read NLA chain
             if self.nla_map:
+                self.buf.seek(self.msg_align(self.buf.tell()))
                 self.decode_nlas()
         except Exception as e:
+            logging.warning(traceback.format_exc())
             raise NetlinkNLADecodeError(e)
         if len(self['attrs']) == 0:
             del self['attrs']
@@ -834,8 +982,17 @@ class nlmsg_base(dict):
         if self.header is not None:
             self['header'].reserve()
 
-        if self.getvalue() is not None:
-
+        # handle the array case
+        if self.nla_array:
+            for value in self.getvalue():
+                cell = type(self)(self.buf, parent=self, debug=self.debug)
+                cell.nla_array = False
+                if cell.cell_header is not None:
+                    cell.header = cell.cell_header
+                    cell['header'] = cell.cell_header(self.buf)
+                cell.setvalue(value)
+                cell.encode()
+        elif self.getvalue() is not None:
             payload = b''
             for i in self.fields:
                 name = i[0]
@@ -867,12 +1024,13 @@ class nlmsg_base(dict):
                     else:
                         payload += struct.pack(fmt, value)
                 except struct.error:
+                    logging.error(''.join(traceback.format_stack()))
                     logging.error(traceback.format_exc())
                     logging.error("error pack: %s %s %s" %
                                   (fmt, value, type(value)))
                     raise
 
-            diff = NLMSG_ALIGN(len(payload)) - len(payload)
+            diff = self.msg_align(len(payload)) - len(payload)
             self.buf.write(payload)
             self.buf.write(b'\0' * diff)
         # write NLA chain
@@ -894,6 +1052,10 @@ class nlmsg_base(dict):
         if isinstance(value, dict):
             self.update(value)
         else:
+            try:
+                value = self.r_value_map.get(value, value)
+            except TypeError:
+                pass
             self['value'] = value
             self.value = value
 
@@ -924,6 +1086,44 @@ class nlmsg_base(dict):
                    'encoded': 2}
         return [i[fmt_map[fmt]] for i in self['attrs'] if i[0] == attr]
 
+    def load(self, dump):
+        '''
+        Load packet from a structure
+        '''
+        if isinstance(dump, dict):
+            for (k, v) in dump.items():
+                if k == 'header':
+                    self['header'].load(dump['header'])
+                else:
+                    self[k] = v
+        else:
+            self.setvalue(dump)
+
+    def dump(self):
+        '''
+        Dump packet as a simple types struct
+        '''
+        a = self.getvalue()
+        if isinstance(a, dict):
+            ret = {}
+            for (k, v) in a.items():
+                if k == 'header':
+                    ret['header'] = a['header'].dump()
+                elif k == 'attrs':
+                    ret['attrs'] = attrs = []
+                    for i in a['attrs']:
+                        if isinstance(i[1], nlmsg_base):
+                            attrs.append([i[0], i[1].dump()])
+                        elif isinstance(i[1], set):
+                            attrs.append([i[0], tuple(i[1])])
+                        else:
+                            attrs.append([i[0], i[1]])
+                else:
+                    ret[k] = v
+        else:
+            ret = a
+        return ret
+
     def getvalue(self):
         '''
         Atomic NLAs return their value in the 'value' field,
@@ -935,7 +1135,7 @@ class nlmsg_base(dict):
 
         if 'value' in self and self['value'] != NotInitialized:
             # raw value got by generic decoder
-            return self['value']
+            return self.value_map.get(self['value'], self['value'])
 
         return self
 
@@ -959,6 +1159,13 @@ class nlmsg_base(dict):
                          'TCA_HTB_PARMS': (<class 'pyroute2...htb_parms'>, 1),
                          'TCA_HTB_INIT': (<class 'pyroute2...htb_glob'>, 2)}
 
+        nla_map format::
+
+            nla_map = (([ID, ] NAME, TYPE[, FLAGS]), ...)
+
+        Items in `[...]` are optional. If ID is not given, then the map will
+        be autonumerated from 0. If flags are not given, they are 0 by default.
+
         '''
         # clean up NLA mappings
         self.t_nla_map = {}
@@ -968,24 +1175,49 @@ class nlmsg_base(dict):
         if not self.nla_map:
             return
 
+        # fix nla flags
+        nla_map = []
+        for item in self.nla_map:
+            if not isinstance(item[-1], int):
+                item = list(item)
+                item.append(0)
+            nla_map.append(item)
+
         # detect, whether we have pre-defined keys
-        if len(self.nla_map[0]) == 2:
+        if not isinstance(nla_map[0][0], int):
             # create enumeration
-            nla_types = enumerate((i[0] for i in self.nla_map))
+            nla_types = enumerate((i[0] for i in nla_map))
             # that's a little bit tricky, but to reduce
             # the required amount of code in modules, we have
             # to jump over the head
-            zipped = [(k[1][0], k[0][0], k[0][1]) for k in
-                      zip(self.nla_map, nla_types)]
+            zipped = [(k[1][0], k[0][0], k[0][1], k[0][2]) for k in
+                      zip(nla_map, nla_types)]
         else:
-            zipped = self.nla_map
+            zipped = nla_map
 
-        for (key, name, nla_class) in zipped:
+        for (key, name, nla_class, nla_flags) in zipped:
+            # it is an array
+            if nla_class[0] == '*':
+                nla_class = nla_class[1:]
+                nla_array = True
+            else:
+                nla_array = False
+            # are there any init call in the string?
+            lb = nla_class.find('(')
+            rb = nla_class.find(')')
+            if 0 < lb < rb:
+                init = nla_class[lb + 1:rb]
+                nla_class = nla_class[:lb]
+            else:
+                init = None
             # lookup NLA class
-            nla_class = getattr(self, nla_class)
+            if nla_class == 'recursive':
+                nla_class = type(self)
+            else:
+                nla_class = getattr(self, nla_class)
             # update mappings
-            self.t_nla_map[key] = (nla_class, name)
-            self.r_nla_map[name] = (nla_class, key)
+            self.t_nla_map[key] = (nla_class, name, nla_flags, nla_array, init)
+            self.r_nla_map[name] = (nla_class, key, nla_flags, nla_array, init)
 
     def encode_nlas(self):
         '''
@@ -996,13 +1228,17 @@ class nlmsg_base(dict):
             if i[0] in self.r_nla_map:
                 msg_class = self.r_nla_map[i[0]][0]
                 msg_type = self.r_nla_map[i[0]][1]
+                msg_array = self.r_nla_map[i[0]][3]
+                msg_init = self.r_nla_map[i[0]][4]
                 # is it a class or a function?
                 if isinstance(msg_class, types.MethodType):
                     # if it is a function -- use it to get the class
                     msg_class = msg_class()
                 # encode NLA
-                nla = msg_class(self.buf, parent=self)
-                nla['header']['type'] = msg_type
+                nla = msg_class(self.buf, parent=self, init=msg_init)
+                nla.nla_flags |= self.r_nla_map[i[0]][2]
+                nla.nla_array = msg_array
+                nla['header']['type'] = msg_type | nla.nla_flags
                 nla.setvalue(i[1])
                 try:
                     nla.encode()
@@ -1024,6 +1260,8 @@ class nlmsg_base(dict):
             nla = None
             # pick the length and the type
             (length, msg_type) = struct.unpack('HH', self.buf.read(4))
+            # first two bits of msg_type are flags:
+            msg_type = msg_type & ~(NLA_F_NESTED | NLA_F_NET_BYTEORDER)
             # rewind to the beginning
             self.buf.seek(init)
             length = min(max(length, 4),
@@ -1039,22 +1277,35 @@ class nlmsg_base(dict):
                     msg_class = msg_class(buf=self.buf, length=length)
                 # and the name
                 msg_name = self.t_nla_map[msg_type][1]
+                # is it an array?
+                msg_array = self.t_nla_map[msg_type][3]
+                # initstring
+                msg_init = self.t_nla_map[msg_type][4]
                 # decode NLA
-                nla = msg_class(self.buf, length, self, debug=self.debug)
-            try:
-                nla.decode()
-            except:
-                # FIXME
-                self.buf.seek(init)
+                nla = msg_class(self.buf, length, self,
+                                debug=self.debug,
+                                init=msg_init)
+                nla.nla_array = msg_array
+                try:
+                    nla.decode()
+                    nla.nla_flags = msg_type & (NLA_F_NESTED |
+                                                NLA_F_NET_BYTEORDER)
+                except Exception:
+                    logging.warning("decoding %s" % (msg_name))
+                    logging.warning(traceback.format_exc())
+                    self.buf.seek(init)
+                    msg_name = 'UNDECODED'
+                    msg_value = hexdump(self.buf.read(length))
+                else:
+                    msg_value = nla.getvalue()
+            else:
                 msg_name = 'UNKNOWN'
                 msg_value = hexdump(self.buf.read(length))
-            else:
-                msg_value = nla.getvalue()
 
             self['attrs'].append([msg_name, msg_value])
 
             # fix the offset
-            self.buf.seek(init + NLMSG_ALIGN(length))
+            self.buf.seek(init + self.msg_align(length))
 
 
 class nla_header(nlmsg_base):
@@ -1096,6 +1347,16 @@ class nlmsg_atoms(nlmsg_base):
             nla_base.decode(self)
             self.value = None
 
+    class flag(nla_base):
+        '''
+        'flag' type is used to denote attrs that have no payload
+        '''
+        fields = []
+
+        def decode(self):
+            nla_base.decode(self)
+            self.value = True
+
     class uint8(nla_base):
         fields = [('value', 'B')]
 
@@ -1108,6 +1369,9 @@ class nlmsg_atoms(nlmsg_base):
     class uint64(nla_base):
         fields = [('value', 'Q')]
 
+    class int32(nla_base):
+        fields = [('value', 'i')]
+
     class be8(nla_base):
         fields = [('value', '>B')]
 
@@ -1117,31 +1381,32 @@ class nlmsg_atoms(nlmsg_base):
     class be32(nla_base):
         fields = [('value', '>I')]
 
+    class be64(nla_base):
+        fields = [('value', '>Q')]
+
     class ipXaddr(nla_base):
         fields = [('value', 's')]
         family = None
 
         def encode(self):
-            self['value'] = socket.inet_pton(self.family,
-                                             self.value)
+            self['value'] = inet_pton(self.family, self.value)
             nla_base.encode(self)
 
         def decode(self):
             nla_base.decode(self)
-            self.value = socket.inet_ntop(self.family,
-                                          self['value'])
+            self.value = inet_ntop(self.family, self['value'])
 
     class ip4addr(ipXaddr):
         '''
         Explicit IPv4 address type class.
         '''
-        family = socket.AF_INET
+        family = AF_INET
 
     class ip6addr(ipXaddr):
         '''
         Explicit IPv6 address type class.
         '''
-        family = socket.AF_INET6
+        family = AF_INET6
 
     class ipaddr(nla_base):
         '''
@@ -1155,12 +1420,82 @@ class nlmsg_atoms(nlmsg_base):
         fields = [('value', 's')]
 
         def encode(self):
-            self['value'] = socket.inet_pton(self.parent['family'], self.value)
+            # use real provided family, not implicit
+            if self.value.find(':') > -1:
+                family = AF_INET6
+            else:
+                family = AF_INET
+            self['value'] = inet_pton(family, self.value)
             nla_base.encode(self)
 
         def decode(self):
             nla_base.decode(self)
-            self.value = socket.inet_ntop(self.parent['family'], self['value'])
+            # use real provided family, not implicit
+            if self.length > 8:
+                family = AF_INET6
+            else:
+                family = AF_INET
+            self.value = inet_ntop(family, self['value'])
+
+    class target(nla_base):
+        '''
+        A universal target class. The target type depends on the msg
+        family:
+
+        * AF_INET: IPv4 addr, string: "127.0.0.1"
+        * AF_INET6: IPv6 addr, string: "::1"
+        * AF_MPLS: MPLS labels, 0 .. k: [{"label": 0x20, "ttl": 16}, ...]
+        '''
+        fields = [('value', 's')]
+
+        def get_family(self):
+            pointer = self
+            while pointer.parent is not None:
+                pointer = pointer.parent
+            return pointer.get('family', AF_UNSPEC)
+
+        def encode(self):
+            family = self.get_family()
+            if family in (AF_INET, AF_INET6):
+                self['value'] = inet_pton(family, self.value)
+            elif family == AF_MPLS:
+                self['value'] = b''
+                if isinstance(self.value, (set, list, tuple)):
+                    labels = self.value
+                else:
+                    if 'label' in self:
+                        labels = [{'label': self.get('label', 0),
+                                   'tc': self.get('tc', 0),
+                                   'bos': self.get('bos', 0),
+                                   'ttl': self.get('ttl', 0)}]
+                    else:
+                        labels = []
+                for record in labels:
+                    label = (record.get('label', 0) << 12) |\
+                        (record.get('tc', 0) << 9) |\
+                        ((1 if record.get('bos') else 0) << 8) |\
+                        record.get('ttl', 0)
+                    self['value'] += struct.pack('>I', label)
+            else:
+                raise TypeError('socket family not supported')
+            nla_base.encode(self)
+
+        def decode(self):
+            nla_base.decode(self)
+            family = self.get_family()
+            if family in (AF_INET, AF_INET6):
+                self.value = inet_ntop(family, self['value'])
+            elif family == AF_MPLS:
+                self.value = []
+                for i in range(len(self['value']) // 4):
+                    label = struct.unpack('>I', self['value'][i*4:i*4+4])[0]
+                    record = {'label': (label & 0xFFFFF000) >> 12,
+                              'tc': (label & 0x00000E00) >> 9,
+                              'bos': (label & 0x00000100) >> 8,
+                              'ttl': label & 0x000000FF}
+                    self.value.append(record)
+            else:
+                raise TypeError('socket family not supported')
 
     class l2addr(nla_base):
         '''
@@ -1189,18 +1524,52 @@ class nlmsg_atoms(nlmsg_base):
             nla_base.decode(self)
             self.value = hexdump(self['value'])
 
+    class array(nla_base):
+        '''
+        Array of simple data type
+        '''
+        fields = [('value', 's')]
+        _fmt = None
+
+        @property
+        def fmt(self):
+            # try to get format from parent
+            # work only with elementary types
+            if self._fmt is not None:
+                return self._fmt
+            try:
+                fclass = getattr(self.parent, self.nla_init)
+                self._fmt = fclass.fields[0][1]
+            except Exception:
+                self._fmt = self.nla_init
+            return self._fmt
+
+        def encode(self):
+            fmt = '%s%i%s' % (self.fmt[:-1], len(self.value), self.fmt[-1:])
+            self['value'] = struct.pack(fmt, self.value)
+            nla_base.encode(self)
+
+        def decode(self):
+            nla_base.decode(self)
+            data_length = len(self['value'])
+            element_size = struct.calcsize(self.fmt)
+            array_size = data_length // element_size
+            trail = (data_length % element_size) or -data_length
+            data = self['value'][:-trail]
+            fmt = '%s%i%s' % (self.fmt[:-1], array_size, self.fmt[-1:])
+            self.value = struct.unpack(fmt, data)
+
     class cdata(nla_base):
         '''
         Binary data
         '''
         fields = [('value', 's')]
 
-    class asciiz(nla_base):
+    class string(nla_base):
         '''
-        Zero-terminated string.
+        UTF-8 string.
         '''
-        # FIXME: move z-string hacks from general decode here?
-        fields = [('value', 'z')]
+        fields = [('value', 's')]
 
         def encode(self):
             if isinstance(self['value'], str) and sys.version[0] == '3':
@@ -1209,11 +1578,26 @@ class nlmsg_atoms(nlmsg_base):
 
         def decode(self):
             nla_base.decode(self)
-            try:
-                assert sys.version[0] == '3'
-                self.value = self['value'].decode('utf-8')
-            except (AssertionError, UnicodeDecodeError):
-                self.value = self['value']
+            self.value = self['value']
+            if sys.version_info[0] >= 3:
+                try:
+                    self.value = self.value.decode('utf-8')
+                except UnicodeDecodeError:
+                    pass  # Failed to decode, keep undecoded value
+
+    class asciiz(string):
+        '''
+        Zero-terminated string.
+        '''
+        # FIXME: move z-string hacks from general decode here?
+        fields = [('value', 'z')]
+
+    # FIXME: support NLA_FLAG and NLA_MSECS as well.
+    #
+    # aliases to support standard kernel attributes:
+    #
+    binary = cdata       # NLA_BINARY
+    nul_string = asciiz  # NLA_NUL_STRING
 
 
 class nla(nla_base, nlmsg_atoms):
@@ -1250,8 +1634,18 @@ class ctrlmsg(genlmsg):
     nla_map = (('CTRL_ATTR_UNSPEC', 'none'),
                ('CTRL_ATTR_FAMILY_ID', 'uint16'),
                ('CTRL_ATTR_FAMILY_NAME', 'asciiz'),
-               ('CTRL_ATTR_VERSION', 'hex'),
-               ('CTRL_ATTR_HDRSIZE', 'hex'),
-               ('CTRL_ATTR_MAXATTR', 'hex'),
-               ('CTRL_ATTR_OPS', 'hex'),
-               ('CTRL_ATTR_MCAST_GROUPS', 'hex'))
+               ('CTRL_ATTR_VERSION', 'uint32'),
+               ('CTRL_ATTR_HDRSIZE', 'uint32'),
+               ('CTRL_ATTR_MAXATTR', 'uint32'),
+               ('CTRL_ATTR_OPS', '*ops'),
+               ('CTRL_ATTR_MCAST_GROUPS', '*mcast_groups'))
+
+    class ops(nla):
+        nla_map = (('CTRL_ATTR_OP_UNSPEC', 'none'),
+                   ('CTRL_ATTR_OP_ID', 'uint32'),
+                   ('CTRL_ATTR_OP_FLAGS', 'uint32'))
+
+    class mcast_groups(nla):
+        nla_map = (('CTRL_ATTR_MCAST_GRP_UNSPEC', 'none'),
+                   ('CTRL_ATTR_MCAST_GRP_NAME', 'asciiz'),
+                   ('CTRL_ATTR_MCAST_GRP_ID', 'uint32'))
