@@ -1,3 +1,4 @@
+# -*- coding: utf-8 -*-
 '''
 Generic netlink
 ===============
@@ -8,6 +9,9 @@ Describe
 from pyroute2.netlink import CTRL_CMD_GETFAMILY
 from pyroute2.netlink import GENL_ID_CTRL
 from pyroute2.netlink import NLM_F_REQUEST
+from pyroute2.netlink import SOL_NETLINK
+from pyroute2.netlink import NETLINK_ADD_MEMBERSHIP
+from pyroute2.netlink import NETLINK_DROP_MEMBERSHIP
 from pyroute2.netlink import ctrlmsg
 from pyroute2.netlink.nlsocket import NetlinkSocket
 
@@ -17,13 +21,11 @@ class GenericNetlinkSocket(NetlinkSocket):
     Low-level socket interface. Provides all the
     usual socket does, can be used in poll/select,
     doesn't create any implicit threads.
-
-    Provides two additional methods:
-    * get_protocol_id() -- resolve generic netlink proto
-    * get() -- receive and parse netlink messages
     '''
 
-    def bind(self, proto, msg_class, groups=0, pid=0, async=False):
+    mcast_groups = {}
+
+    def bind(self, proto, msg_class, groups=0, pid=None, async=False):
         '''
         Bind the socket and performs generic netlink
         proto lookup. The `proto` parameter is a string,
@@ -32,13 +34,28 @@ class GenericNetlinkSocket(NetlinkSocket):
         '''
         NetlinkSocket.bind(self, groups, pid, async)
         self.marshal.msg_map[GENL_ID_CTRL] = ctrlmsg
-        self.prid = self.get_protocol_id(proto)
+        msg = self.discovery(proto)
+        self.prid = msg.get_attr('CTRL_ATTR_FAMILY_ID')
+        self.mcast_groups = \
+            dict([(x.get_attr('CTRL_ATTR_MCAST_GRP_NAME'),
+                   x.get_attr('CTRL_ATTR_MCAST_GRP_ID')) for x
+                  in msg.get_attr('CTRL_ATTR_MCAST_GROUPS', [])])
         self.marshal.msg_map[self.prid] = msg_class
 
-    def get_protocol_id(self, proto):
+    def add_membership(self, group):
+        self.setsockopt(SOL_NETLINK,
+                        NETLINK_ADD_MEMBERSHIP,
+                        self.mcast_groups[group])
+
+    def drop_membership(self, group):
+        self.setsockopt(SOL_NETLINK,
+                        NETLINK_DROP_MEMBERSHIP,
+                        self.mcast_groups[group])
+
+    def discovery(self, proto):
         '''
         Resolve generic netlink protocol -- takes a string
-        as the only parameter, return integer proto ID
+        as the only parameter, return protocol description
         '''
         msg = ctrlmsg()
         msg['cmd'] = CTRL_CMD_GETFAMILY
@@ -53,4 +70,4 @@ class GenericNetlinkSocket(NetlinkSocket):
         err = msg['header'].get('error', None)
         if err is not None:
             raise err
-        return msg.get_attr('CTRL_ATTR_FAMILY_ID')
+        return msg
