@@ -1,28 +1,51 @@
-
+from pyroute2.common import map_namespace
 from pyroute2.netlink import nlmsg
 from pyroute2.netlink import nla
 
 
+NUD_INCOMPLETE = 0x01
+NUD_REACHABLE = 0x02
+NUD_STALE = 0x04
+NUD_DELAY = 0x08
+NUD_PROBE = 0x10
+NUD_FAILED = 0x20
+
+# dummy states
+
+NUD_NOARP = 0x40
+NUD_PERMANENT = 0x80
+NUD_NONE = 0x00
+
+(NUD_NAMES, NUD_VALUES) = map_namespace('NUD_', globals())
+
+
 class ndmsg(nlmsg):
     '''
     ARP cache update message
 
-    struct ndmsg {
-        unsigned char ndm_family;
-        int           ndm_ifindex;  /* Interface index */
-        __u16         ndm_state;    /* State */
-        __u8          ndm_flags;    /* Flags */
-        __u8          ndm_type;
-    };
+    C structure::
 
-    struct nda_cacheinfo {
-        __u32         ndm_confirmed;
-        __u32         ndm_used;
-        __u32         ndm_updated;
-        __u32         ndm_refcnt;
-    };
+        struct ndmsg {
+            unsigned char ndm_family;
+            int           ndm_ifindex;  /* Interface index */
+            __u16         ndm_state;    /* State */
+            __u8          ndm_flags;    /* Flags */
+            __u8          ndm_type;
+        };
+
+    Cache info structure::
+
+        struct nda_cacheinfo {
+            __u32         ndm_confirmed;
+            __u32         ndm_used;
+            __u32         ndm_updated;
+            __u32         ndm_refcnt;
+        };
     '''
-    fields = (('family', 'i'),
+    prefix = 'NDA_'
+
+    fields = (('family', 'B'),
+              ('__pad', '3x'),
               ('ifindex', 'i'),
               ('state', 'H'),
               ('flags', 'B'),
@@ -46,7 +69,8 @@ class ndmsg(nlmsg):
                ('NDA_VLAN', 'uint16'),
                ('NDA_PORT', 'be16'),
                ('NDA_VNI', 'be32'),
-               ('NDA_IFINDEX', 'uint32'))
+               ('NDA_IFINDEX', 'uint32'),
+               ('NDA_MASTER', 'uint32'))
 
     class cacheinfo(nla):
         fields = (('ndm_confirmed', 'I'),
